@@ -7,20 +7,21 @@ TODO:
 - [ ] Load HGPS maps
 - [ ] Source object should contain info on components and associations
 - [ ] Links to SNRCat
-- [ ] Show image in ds9 or js9
 """
 from __future__ import absolute_import, division, print_function, unicode_literals
 import os
 from collections import OrderedDict
 import numpy as np
-from astropy.tests.helper import ignore_warnings
 import astropy.units as u
 from astropy.table import Table
 from astropy.coordinates import Angle
+from astropy.modeling.models import Gaussian2D
 from ..extern.pathlib import Path
 from ..utils.scripts import make_path
 from ..spectrum import FluxPoints
 from ..spectrum.models import PowerLaw, ExponentialCutoffPowerLaw
+from ..image.models import Delta2D, Shell2D
+from ..background.models import GaussianBand2D
 from .core import SourceCatalog, SourceCatalogObject
 
 __all__ = [
@@ -47,17 +48,6 @@ class HGPSGaussComponent(object):
     def __init__(self, data):
         self.data = OrderedDict(data)
 
-    @property
-    def name(self):
-        """Source name"""
-        name = self.data[self._source_name_key]
-        return name.strip()
-
-    @property
-    def index(self):
-        """Row index of source in catalog"""
-        return self.data[self._source_index_key]
-
     def __str__(self):
         """Pretty-print source data"""
         d = self.data
@@ -72,26 +62,44 @@ class HGPSGaussComponent(object):
         ss += fmt.format('Flux (>1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB)
         return ss
 
+    @property
+    def name(self):
+        """Source name"""
+        name = self.data[self._source_name_key]
+        return name.strip()
+
+    @property
+    def index(self):
+        """Row index of source in catalog"""
+        return self.data[self._source_index_key]
+
+    @property
+    def spatial_model(self):
+        """
+        Component spatial model.
+        """
+        d = self.data
+        # At the moment spatial models are normalised by deg^2
+        amplitude = d['Flux_Map'].value / (2 * np.pi * d['Size'].value ** 2)
+        glon = Angle(d['GLON']).wrap_at('180d')
+        glat = Angle(d['GLAT']).wrap_at('180d')
+
+        return Gaussian2D(
+            x_mean=glon.value,
+            y_mean=glat.value,
+            x_stddev=d['Size'].value,
+            y_stddev=d['Size'].value,
+            amplitude=amplitude,
+        )
+
 
 class SourceCatalogObjectHGPS(SourceCatalogObject):
     """One object from the HGPS catalog."""
 
-    @property
-    def energy_range(self):
-        return u.Quantity([self.data['Energy_Range_Spec_Lo'], self.data['Energy_Range_Spec_Hi']])
+    def __str__(self):
+        return self.info()
 
     def info(self, info='all'):
-        """Print info.
-
-        Parameters
-        ----------
-        info : {'all', 'basic', 'map', 'spec', 'flux_points', 'components', 'associations'}
-            Comma separated list of options
-        """
-        ss = self.__str__(info=info)
-        print(ss)
-
-    def __str__(self, info='all'):
         """Info string.
 
         Parameters
@@ -100,7 +108,7 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
             Comma separated list of options
         """
         if info == 'all':
-            info = 'basic,map,spec,flux_points,components,associations'
+            info = 'basic,associations,map,spec,flux_points,components'
 
         ss = ''
         ops = info.split(',')
@@ -132,6 +140,12 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         ss += '\n'
         return ss
 
+    def _info_associations(self):
+        ss = '\n*** Source associations info ***\n\n'
+        associations = ', '.join(self.associations)
+        ss += 'List of associated objects: {}\n'.format(associations)
+        return ss
+
     def _info_map(self):
         """Print info from map analysis."""
         d = self.data
@@ -168,30 +182,43 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
 
         ss += '{:<20s} : {:.1f} hours\n'.format('Livetime', d['Livetime'].value)
 
-        ss += '{:<20s} : {:.1f}\n'.format('Energy threshold', d['Energy_Threshold'])
+        ss += '{:<20s} : {:.2f}\n'.format('Energy threshold', d['Energy_Threshold'])
 
         val, err = d['Flux_Map'].value, d['Flux_Map_Err'].value
-        ss += '{:<20s} : ({:.2f} +/- {:.2f}) x 10^-12 cm^-2 s^-1 = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'Source flux (>1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB)
 
         ss += '\nFluxes in RSpec (> 1 TeV):\n'
 
-        ss += '{:<30s} : {:.2f} x 10^-12 cm^-2 s^-1 = {:5.1f} % Crab\n'.format(
-            'Map measurement', d['Flux_Map_RSpec_Data'].value / FF, d['Flux_Map_RSpec_Data'].value * FLUX_TO_CRAB)
+        ss += '{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n'.format(
+            'Map measurement',
+            d['Flux_Map_RSpec_Data'].value / FF,
+            d['Flux_Map_RSpec_Data'].value * FLUX_TO_CRAB,
+        )
 
-        ss += '{:<30s} : {:.2f} x 10^-12 cm^-2 s^-1 = {:5.1f} % Crab\n'.format(
-            'Source model', d['Flux_Map_RSpec_Source'].value / FF, d['Flux_Map_RSpec_Source'].value * FLUX_TO_CRAB)
+        ss += '{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n'.format(
+            'Source model',
+            d['Flux_Map_RSpec_Source'].value / FF,
+            d['Flux_Map_RSpec_Source'].value * FLUX_TO_CRAB,
+        )
 
-        ss += '{:<30s} : {:.2f} x 10^-12 cm^-2 s^-1 = {:5.1f} % Crab\n'.format(
-            'Other component model', d['Flux_Map_RSpec_Other'].value / FF,
-                                     d['Flux_Map_RSpec_Other'].value * FLUX_TO_CRAB)
+        ss += '{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n'.format(
+            'Other component model',
+            d['Flux_Map_RSpec_Other'].value / FF,
+            d['Flux_Map_RSpec_Other'].value * FLUX_TO_CRAB,
+        )
 
-        ss += '{:<30s} : {:.2f} x 10^-12 cm^-2 s^-1 = {:5.1f} % Crab\n'.format(
-            'Large scale component model', d['Flux_Map_RSpec_LS'].value / FF,
-                                           d['Flux_Map_RSpec_LS'].value * FLUX_TO_CRAB)
+        ss += '{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n'.format(
+            'Large scale component model',
+            d['Flux_Map_RSpec_LS'].value / FF,
+            d['Flux_Map_RSpec_LS'].value * FLUX_TO_CRAB,
+        )
 
-        ss += '{:<30s} : {:.2f} x 10^-12 cm^-2 s^-1 = {:5.1f} % Crab\n'.format(
-            'Total model', d['Flux_Map_RSpec_Total'].value / FF, d['Flux_Map_RSpec_Total'].value * FLUX_TO_CRAB)
+        ss += '{:<30s} : {:.3f} x 10^-12 cm^-2 s^-1 = {:5.2f} % Crab\n'.format(
+            'Total model',
+            d['Flux_Map_RSpec_Total'].value / FF,
+            d['Flux_Map_RSpec_Total'].value * FLUX_TO_CRAB,
+        )
 
         ss += '{:<35s} : {:5.1f} %\n'.format('Containment in RSpec', 100 * d['Containment_RSpec'])
         ss += '{:<35s} : {:5.1f} %\n'.format('Contamination in RSpec', 100 * d['Contamination_RSpec'])
@@ -209,9 +236,9 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
 
         ss += '{:<20s} : {:.1f} hours\n'.format('Livetime', d['Livetime_Spec'].value)
 
-        lo = d['Energy_Range_Spec_Lo'].value
-        hi = d['Energy_Range_Spec_Hi'].value
-        ss += '{:<20s} : {:.1f} to {:.1f} TeV\n'.format('Energy range:', lo, hi)
+        lo = d['Energy_Range_Spec_Min'].value
+        hi = d['Energy_Range_Spec_Max'].value
+        ss += '{:<20s} : {:.2f} to {:.2f} TeV\n'.format('Energy range:', lo, hi)
 
         ss += '{:<20s} : {:.1f}\n'.format('Background', d['Background_Spec'])
         ss += '{:<20s} : {:.1f}\n'.format('Excess', d['Excess_Spec'])
@@ -222,12 +249,12 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
 
         val = d['Flux_Spec_Int_1TeV'].value
         err = d['Flux_Spec_Int_1TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'Best-fit model flux(> 1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB)
 
         val = d['Flux_Spec_Energy_1_10_TeV'].value
         err = d['Flux_Spec_Energy_1_10_TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 erg cm^-2 s^-1\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 erg cm^-2 s^-1\n'.format(
             'Best-fit model energy flux(1 to 10 TeV)', val / FF, err / FF)
 
         # TODO: can we just use the Gammapy model classes here instead of duplicating the code?
@@ -238,21 +265,21 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
 
     def _info_spec_pl(self):
         d = self.data
-        ss = '{:<20s} : {:.1f}\n'.format('Pivot energy', d['Energy_Spec_PL_Pivot'])
+        ss = '{:<20s} : {:.2f}\n'.format('Pivot energy', d['Energy_Spec_PL_Pivot'])
 
         val = d['Flux_Spec_PL_Diff_Pivot'].value
         err = d['Flux_Spec_PL_Diff_Pivot_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'Flux at pivot energy', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB_DIFF)
 
         val = d['Flux_Spec_PL_Int_1TeV'].value
         err = d['Flux_Spec_PL_Int_1TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'PL   Flux(> 1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB)
 
         val = d['Flux_Spec_PL_Diff_1TeV'].value
         err = d['Flux_Spec_PL_Diff_1TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'PL   Flux(@ 1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB_DIFF)
 
         val = d['Index_Spec_PL']
@@ -268,12 +295,12 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
 
         val = d['Flux_Spec_ECPL_Diff_1TeV'].value
         err = d['Flux_Spec_ECPL_Diff_1TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1 TeV^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'ECPL   Flux(@ 1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB_DIFF)
 
         val = d['Flux_Spec_ECPL_Int_1TeV'].value
         err = d['Flux_Spec_ECPL_Int_1TeV_Err'].value
-        ss += '{:<20s} : ({:.1f} +/- {:.1f}) x 10^-12 cm^-2 s^-1  = ({:.1f} +/- {:.1f}) % Crab\n'.format(
+        ss += '{:<20s} : ({:.3f} +/- {:.3f}) x 10^-12 cm^-2 s^-1  = ({:.2f} +/- {:.2f}) % Crab\n'.format(
             'ECPL   Flux(> 1 TeV)', val / FF, err / FF, val * FLUX_TO_CRAB, err * FLUX_TO_CRAB)
 
         val = d['Index_Spec_ECPL']
@@ -289,7 +316,7 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         err = err / val ** 2
         val = 1. / val
 
-        ss += '{:<20s} : {:.1f} +/- {:.1f} TeV\n'.format('ECPL E_cut', val, err)
+        ss += '{:<20s} : {:.2f} +/- {:.2f} TeV\n'.format('ECPL E_cut', val, err)
 
         return ss
 
@@ -303,8 +330,9 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         flux_points = self.flux_points.table.copy()
 
         energy_cols = ['e_ref', 'e_min', 'e_max']
-        flux_cols = ['dnde', 'dnde_errn', 'dnde_errp']
-        flux_points = flux_points[energy_cols + flux_cols]
+        flux_cols = ['dnde', 'dnde_errn', 'dnde_errp', 'dnde_ul']
+        cols = energy_cols + flux_cols + ['is_ul']
+        flux_points = flux_points[cols]
 
         for _ in energy_cols:
             flux_points[_].format = '.3f'
@@ -313,7 +341,8 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
             flux_points[_].format = '.3e'
 
         # convert table to string
-        ss += '\n\t'.join(flux_points.pformat(-1))
+        lines = flux_points.pformat(max_width=-1, max_lines=-1)
+        ss += '\n'.join(lines)
         return ss + '\n'
 
     def _info_components(self):
@@ -326,16 +355,25 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         ss += '{:<20s} : {}\n\n'.format('Spatial components', self.data['Components'])
 
         for component in self.components:
-            # Call __str__ directly to
-            ss += component.__str__()
+            ss += str(component)
             ss += '\n\n'
+
         return ss
 
-    def _info_associations(self):
-        ss = '\n*** Source associations info ***\n\n'
-        associations = ', '.join(self.associations)
-        ss += 'List of associated objects: {}\n'.format(associations)
-        return ss
+    @property
+    def energy_range(self):
+        """Spectral model energy range (`~astropy.units.Quantity` with length 2)."""
+        e = u.Quantity([
+            self.data['Energy_Range_Spec_Min'],
+            self.data['Energy_Range_Spec_Max'],
+        ])
+        # Some EXTERN sources have no energy range information.
+        # In those cases, we put a default
+        use_default = np.isnan(e)
+        e_default = [0.2, 50] * u.TeV
+        e[use_default] = e_default[use_default]
+
+        return e
 
     @property
     def spectral_model(self):
@@ -348,6 +386,23 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         """
         data = self.data
         spec_type = data['Spectral_Model'].strip()
+        return self._spectral_model(spec_type)
+
+    def _spectral_model(self, spec_type):
+        """
+        Create spectral model from source parameters.
+
+        Parameters
+        ----------
+        spec_type : str
+            Spectral model type either 'PL' or 'ECPL'.
+
+        Returns
+        -------
+        model : `~gammapy.spectrum.models.SpectralModel`
+            Spectral model instance.
+        """
+        data = self.data
 
         pars, errs = {}, {}
 
@@ -373,6 +428,81 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         model.parameters.set_parameter_errors(errs)
         return model
 
+    def spatial_model(self, emin=1 * u.TeV, emax=1E5 * u.TeV):
+        """
+        Source spatial model.
+        """
+        d = self.data
+        flux = self.spectral_model.integral(emin, emax)
+        amplitude = flux.to('cm-2 s-1').value
+
+        pars = {}
+        glon = Angle(d['GLON']).wrap_at('180d')
+        glat = Angle(d['GLAT']).wrap_at('180d')
+
+        spatial_model = d['Spatial_Model']
+
+        if self.is_pointlike:
+            pars['amplitude'] = amplitude
+            pars['x_0'] = glon.value
+            pars['y_0'] = glat.value
+            return Delta2D(**pars)
+
+        elif spatial_model == 'Shell':
+            # HGPS contains no information on shell width
+            # Here we assuma a 5% shell width for all shells.
+            shell_width_fraction = 0.05
+            pars['x_0'] = glon.value
+            pars['y_0'] = glat.value
+            r_out = d['Size'].to('deg').value
+            pars['width'] = r_out * shell_width_fraction
+            pars['r_in'] = r_out * (1 - shell_width_fraction)
+            pars['amplitude'] = amplitude
+            return Shell2D(**pars)
+
+        elif spatial_model in ['Gaussian', '2-Gaussian', '3-Gaussian']:
+            amplitude_total = d['Flux_Map'].to('cm-2 s-1').value
+            try:
+                models = [component.spatial_model for component in self.components]
+            except AttributeError:
+                # there is one external source (HESS J1801-233) where there is no
+                # component info available, so we create it here locally
+                models = [HGPSGaussComponent(d).spatial_model]
+
+            for model in models:
+                # weight total flux according to relative amplitude
+                model.amplitude = model.amplitude / amplitude_total * amplitude
+
+            if spatial_model == 'Gaussian':
+                return models[0]
+
+            elif spatial_model == '2-Gaussian':
+                model = models[0] + models[1]
+
+                # use generic bounding box of 2 deg width and height
+                model.bounding_box = ((glat.deg - 2, glat.deg + 2),
+                                      (glon.deg - 2, glon.deg + 2))
+                return model
+
+            elif spatial_model == '3-Gaussian':
+                model = models[0] + models[1] + models[2]
+
+                # use generic bounding box of 2 deg width and height
+                model.bounding_box = ((glat.deg - 2, glat.deg + 2),
+                                      (glon.deg - 2, glon.deg + 2))
+                return model
+
+        else:
+            raise ValueError('Not a valid spatial model{}'.format(spatial_model))
+
+    @property
+    def is_pointlike(self):
+        """Source is pointlike"""
+        d = self.data
+        has_size_ul = np.isfinite(d['Size_UL'])
+        pointlike = d['Spatial_Model'] == 'Point-Like'
+        return pointlike or has_size_ul
+
     @property
     def flux_points(self):
         """Flux points (`~gammapy.spectrum.FluxPoints`)."""
@@ -385,9 +515,10 @@ class SourceCatalogObjectHGPS(SourceCatalogObject):
         table['e_max'] = self.data['Flux_Points_Energy_Max'][mask]
 
         table['dnde'] = self.data['Flux_Points_Flux'][mask]
-        table['dnde_errp'] = self.data['Flux_Points_Flux_Err_Hi'][mask]
         table['dnde_errn'] = self.data['Flux_Points_Flux_Err_Lo'][mask]
+        table['dnde_errp'] = self.data['Flux_Points_Flux_Err_Hi'][mask]
         table['dnde_ul'] = self.data['Flux_Points_Flux_UL'][mask]
+        table['is_ul'] = self.data['Flux_Points_Flux_Is_UL'][mask].astype('bool')
 
         return FluxPoints(table)
 
@@ -407,9 +538,7 @@ class SourceCatalogHGPS(SourceCatalog):
             filename = Path(os.environ['HGPS_ANALYSIS']) / 'data/catalogs/HGPS3/release/HGPS_v0.4.fits'
 
         filename = str(make_path(filename))
-
-        with ignore_warnings():  # ignore FITS units warnings
-            table = Table.read(filename, hdu=hdu)
+        table = Table.read(filename, hdu=hdu)
 
         source_name_alias = ('Identified_Object',)
         super(SourceCatalogHGPS, self).__init__(
@@ -420,6 +549,14 @@ class SourceCatalogHGPS(SourceCatalog):
         self.components = Table.read(filename, hdu='HGPS_GAUSS_COMPONENTS')
         self.associations = Table.read(filename, hdu='HGPS_ASSOCIATIONS')
         self.identifications = Table.read(filename, hdu='HGPS_IDENTIFICATIONS')
+        self._large_scale_component = Table.read(filename, hdu='HGPS_LARGE_SCALE_COMPONENT')
+
+    @property
+    def large_scale_component(self):
+        """Large sclae component model (`~gammapy.background.models.GaussianBand2D`).
+        """
+        table = self._large_scale_component
+        return GaussianBand2D(table, spline_kwargs=dict(k=3, s=0, ext=0))
 
     def _make_source_object(self, index):
         """Make one source object.
