@@ -7,7 +7,8 @@ from astropy.coordinates import SkyCoord, Angle
 from astropy.io import fits
 from astropy import units as u
 from astropy.units import Quantity
-from astropy.tests.helper import pytest, assert_quantity_allclose
+from astropy.tests.helper import assert_quantity_allclose
+import pytest
 from astropy.wcs import WcsError
 from regions import PixCoord, CirclePixelRegion, CircleSkyRegion
 from ...utils.testing import requires_dependency, requires_data
@@ -307,7 +308,7 @@ class TestSkyImage:
         self.image = SkyImage.empty(
             nxpix=10, nypix=5, binsz=0.2,
             xref=self.center.l.deg, yref=self.center.b.deg,
-            proj='TAN', coordsys='GAL',
+            proj='TAN', coordsys='GAL', meta={'TEST': 42},
         )
         self.image.data = np.arange(50, dtype=float).reshape((5, 10))
 
@@ -321,6 +322,14 @@ class TestSkyImage:
         assert_allclose(center.l.deg, self.center.l.deg, atol=1e-5)
         assert_allclose(center.b.deg, self.center.b.deg, atol=1e-5)
 
+    def test_width(self):
+        width = self.image.width
+        assert_quantity_allclose(width, 2 * u.deg, rtol=1e-3)
+
+    def test_height(self):
+        height = self.image.height
+        assert_allclose(height, 1 * u.deg, rtol=1e-3)
+
     @requires_dependency('scipy')
     @pytest.mark.parametrize('kernel', ['gauss', 'box', 'disk'])
     def test_smooth(self, kernel):
@@ -329,6 +338,9 @@ class TestSkyImage:
         actual = smoothed.data.sum()
         assert_allclose(actual, desired)
 
+    def test_meta(self):
+        assert self.image.meta['TEST'] == 42
+
 
 def test_image_pad():
     image = SkyImage.empty(nxpix=10, nypix=13)
@@ -452,3 +464,15 @@ def test_distance_image():
     distance = mask.distance_image.data
     expected = [[-1, -1, 1], [1, 1, 1.41421356]]
     assert_allclose(distance, expected)
+
+
+@requires_dependency('scipy')
+def test_conversion_wcs_map_nd():
+    """Check conversion SkyCube <-> WCSMapNd"""
+    image = SkyImage.empty(nxpix=3, nypix=2, unit='cm', name='axel')
+
+    map = image.to_wcs_nd_map()
+    image2 = SkyImage.from_wcs_nd_map(map)
+
+    # Note: WCSMapNd doesn't store name or unit at the moment
+    SkyImage.assert_allclose(image, image2, check_name=False, check_unit=False)
